# flake8: noqa: I900
class GenericAudienceTypes:
    """
    This class contains the different types of generic audiences.
    .. remarks::

        .. note::
            This class is used to define the types of generic audiences that can be used in the generic_connection decorator.
    """
    CosmosDB = "CosmosDB"
    KeyVault = "KeyVault"